//nxtcp.cpp - command line utility for downloading and uploading files and programs to/from NXT

#include <stdio.h>
#include <string.h>
#include <errno.h>

#ifdef WIN32
#include <windows.h>
#include <io.h>
#define SEARCHHANDLETYPE intptr_t
char slash='\\';
#else
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <malloc.h>
#include <errno.h>
#define HANDLE int
#define INVALID_HANDLE_VALUE -1
#define WriteFile(A,B,C,D,E) write(A,B,C)
#define ReadFile(A,B,C,D,E) read(A,B,C)
#define FlushFileBuffers(A) fsync(A)
#define CloseHandle close
#define DWORD unsigned int
#define intptr_t DIR *
#define SEARCHHANDLETYPE DIR *
#define _finddata32_t dirent *
#define _fileno fileno
#define Sleep(X) usleep(1000*X)
char slash='/';
int _filelength(int f)
{
 struct stat fstatbuf;
 fstat(f, &fstatbuf);
 return fstatbuf.st_size;
}

//  Case-sensitive pattern match
int patMatch (const char* pattern, const char* string)
{
      switch (pattern[0])
      {
      case '\0': return !string[0];
      case '*': return patMatch(pattern+1, string) || string[0] && patMatch(pattern, string+1);
      case '?': return string[0] && patMatch(pattern+1, string+1);
      default: return (pattern[0] == string[0]) && patMatch(pattern+1, string+1);
      }
}

static char *findnextpattern;
struct dirent *_findnext32(DIR *handle, struct dirent **finfo)
{
  do {
    *finfo = readdir(handle);
    if (*finfo)
    {
      if (patMatch(findnextpattern, (*finfo)->d_name)) return *finfo;
    }
  }
  while (*finfo);
  return *finfo;
}

DIR *_findfirst32(char *pattern, struct dirent **finfo)
{
  DIR *handle;
  int l = strlen(pattern);
  if (!l) return 0;

  /* strip the filename, keep only path, but store the filename/pattern to findnextpattern */
  char *q = pattern + l;
  char *p = q - 1;
  while ((p >= pattern) && (*p != '/')) p--;
  p++;
  l = p - pattern;
  
  findnextpattern = (char *)malloc(q - p + 1);
  strncpy(findnextpattern, p, q - p + 1);
  p = (char *)malloc(l + 1);
  strncpy(p, pattern, l);
  p[l] = '\0';

  if (strlen(p) == 0) handle = opendir(".");
  else handle = opendir(p);
  
  free(p);
  if (!handle)
  { 
    perror("Could not open dir");
  }
  _findnext32(handle, finfo);
  return handle;
}

void _findclose(DIR *handle)
{
  closedir(handle);
  free(findnextpattern);
}
#endif

static HANDLE handle;

// send NXT message string to port
void send(unsigned char *data)
{
  DWORD written;
  WriteFile(handle, data, (DWORD)(data[0] + 2), &written, 0);
  FlushFileBuffers(handle);
}

// receive one byte from port (returns after ca. 1.5 seconds if nothing received)
// returns -1 in case of error, -2 in case of timeout, otherwise the received char
int receive(void)
{
  unsigned char buf;
  DWORD nread;
  int rv = ReadFile(handle, &buf, 1, &nread, 0);
#ifdef WIN32
  if (rv == 0) return -1;
  if (nread == 0) return -2;
#else
  if (rv == -1) return -1;
  if (rv == 0) return -2;
#endif
  return buf;
}

void print_usage()
{
	printf("\nnxtcp, copy files from/to NXT over BlueTooth virtual serial port\n");
	printf("\nusage: nxtcp [nxt_port:|src_path%c]filename1 [[nxt_port:|dst_path%c][filename2]]\n", slash, slash);
#ifdef WIN32
	char *port = "COM21";
#else
	char *port = "/dev/rfcomm0";
#endif
	printf("\nexamples: nxtcp %s:data.txt\n", port);
	printf("            - copy file data.txt from NXT to cuurent dir at PC\n\n");
	printf("          nxtcp fib.lgo %s:\n", port);
	printf("            - copy file fib.lgo from PC to NXT\n\n");
	printf("          nxtcp %s:data.txt measure.txt\n", port);
	printf("            - copy file data.txt from NXT to file measure.txt at PC\n\n");
	printf("          nxtcp prog.txt %s:prog.lgo\n", port);
	printf("            - copy file prog.txt from PC to NXT file prog.lgo\n\n");
	printf("          nxtcp %s:*.* backup_dir%c\n", port, slash);
	printf("            - copy all files from NXT to backup_dir on PC\n\n");
	printf("          nxtcp nxt_melodies%c*.* %s:\n", slash, port);
	printf("            - copy all files from specified location to NXT\n\n");
}

char *nxt_error(int e)
{
	static char *nxterrs[] = { "No more handles", "No space", "No more files",
		                       "End of file expected", "End of file", "Not a linear file",
							   "File not found", "Handle all ready closed", "No linear space",
							   "Undefined error", "File is busy", "No write buffers",
							   "Append not possible", "File is full", "File exists",
							   "Module not found", "Out of boundary", "Illegal file name",
							   "Illegal handle" };
	if (!e) return "OK";
	if ((e > 0x80) && (e <= 0x93)) return nxterrs[e - 0x81];
	return "Unknown Error";
}

int main(int argc, char **argv)
{
  unsigned char buf[48];
  unsigned char p[100];
  unsigned char nxt_msg[100];
  char tmpname[400], local_path[400];
  char portname[20];
  char *comarg, *local_filename, *nxt_filename, *local_file, *nxt_file;
  char *tmps, *tmps2, *nm;
  int send_direction, multiple, d, len, i, linear, skip_reply, nread, nwrote, written;
  unsigned char nxthandle;
  struct _finddata32_t finfo;
  intptr_t searchhandle;
  unsigned char ffhandle;

  //extract virtual port name from command line
  if (argc < 2)
  {
    print_usage();
	return 0;
  }

  // first determine the NXT port name (must start with 'COM')
  // NXT filename and local filename with path
  local_filename = 0;
  if ((strncmp(argv[1], "COM", 3) == 0) || (strncmp(argv[1], "com", 3) == 0)
#ifndef WIN32
      || (strncmp(argv[1], "/dev/", 5) == 0)
#endif
      )
  {
    send_direction = 0;
    comarg = argv[1];
    if (argc > 2) 
	{
	  local_filename = argv[2];
	  if ((strncmp(argv[2], "COM", 3) == 0) || (strncmp(argv[2], "com", 3) == 0)
#ifndef WIN32
             || (strncmp(argv[2], "/dev/", 5) == 0)
#endif
)
	  {
	    print_usage();
	    printf(" Only one of the two arguments can be remote NXT files\n");
	    return 0;
	  }
	}
  }
  else if (argc == 2) 
  {
	print_usage();
	printf(" If you supply only one argument, it must be the NXT file name\n\n");
	return 0;
  }
  else if ((strncmp(argv[2], "COM", 3) == 0) || (strncmp(argv[2], "com", 3) == 0)
#ifndef WIN32
     || (strncmp(argv[2], "/dev/", 5) == 0)
#endif
      )
  {
 	comarg = argv[2];
    local_filename = argv[1]; 
	send_direction = 1;
  }
  else 
  {
	print_usage();
	printf(" At least one of the arguments needs to be a remote NXT file,\n");
#ifdef WIN32
	printf(" and the NXT port should have a form of COMX: or COMXX:\n\n");
#else
	printf(" and the NXT port should have a form of /dev/rfcommX:\n\n");
#endif
	return 0;
  }

#ifdef WIN32
  if (comarg[4] == ':') 
  {
    strncpy(portname, comarg, 4);
	portname[4] = 0;
	nxt_filename = comarg + 5;
  }
  else 
  {
    strncpy(portname, "\\\\.\\", 4);
    strncpy(portname + 4, comarg, 5);
	portname[9] = 0;
	nxt_filename = comarg + 6;
  }
#else
  int pos;
  if (comarg[12] == ':')  pos = 12;
  else if (comarg[13] == ':') pos = 13;
  else
  { 
    printf(" I do not understand %s\n", comarg);
    return 0;
  }
  strncpy(portname, comarg, pos);
  portname[pos] = 0;
  nxt_filename = comarg + pos + 1;
#endif

  if (local_filename == 0) local_filename = nxt_filename;
  if ((nxt_filename[0] == 0) && !send_direction) 
  {
	print_usage();
	printf(" You need to specify the file(s) to upload from NXT.\n\n");
	return 0;
  }
  // if no destination nxt filename given, use the same filename as local
  if (nxt_filename[0] == 0) 
  {
	  nxt_filename = local_filename;
      // however, use only the filename, strip the path
	  while ((tmps = strchr(nxt_filename, ':'))) nxt_filename = tmps + 1;
	  while ((tmps = strchr(nxt_filename, '\\'))) nxt_filename = tmps + 1;
	  while ((tmps = strchr(nxt_filename, '/'))) nxt_filename = tmps + 1;
  }

  if (strlen(nxt_filename) > 19)
  {
	  printf("NXT allows only max. 15.3 characters in filename, you supplied: %s\n", nxt_filename);
	  return 1;
  }
  if (strlen(local_filename) > 300)
  {
	  printf("Too long filename %s\n", local_filename);
	  return 1;
  }

  multiple = 0;
  //in case there are wildcards in the local filename, we will do multiple upload
  if (strchr(local_filename, '*') != 0) multiple = 1; 
  //in case there are wildcards in the nxt file name, we will do multiple download
  if (strchr(nxt_filename, '*') != 0) multiple = 1;

  //determine the path of local file
  tmps = local_filename;
  while ((tmps2 = strchr(tmps, '\\'))) tmps = tmps2 + 1;
  while ((tmps2 = strchr(tmps, '/'))) tmps = tmps2 + 1;
  strncpy(local_path, local_filename, tmps - local_filename);
  local_path[tmps - local_filename] = '\0';

  // open the NXT serial port
#ifdef WIN32
  handle = CreateFileA(portname, FILE_READ_DATA | FILE_APPEND_DATA | FILE_WRITE_DATA, 0, NULL, OPEN_EXISTING, 0, NULL);
  if (handle != INVALID_HANDLE_VALUE) 
  {
    SetupComm(handle, 90, 90);

    COMMTIMEOUTS timeouts;
    timeouts.ReadIntervalTimeout = 0;
    timeouts.ReadTotalTimeoutConstant = 0;
    timeouts.ReadTotalTimeoutMultiplier = 1500;  //1.5s timeout on read() from serial port
    timeouts.WriteTotalTimeoutMultiplier = 0;
    timeouts.WriteTotalTimeoutConstant = 1500;  //1.5s timeout on write() to serial port
    SetCommTimeouts(handle, &timeouts);

    PurgeComm(handle, PURGE_RXCLEAR | PURGE_TXCLEAR);
  }
  else 
  {
    printf("Could not open the port %s, err=%d\n\n", portname, GetLastError());
    return 1;
  }
#else
    struct termios newtio, oldtio;
    handle = open(portname, O_RDWR | O_NOCTTY);
    if (handle < 0) { perror(portname); return INVALID_HANDLE_VALUE; }

    tcgetattr(handle,&oldtio); /* save current port settings */

    bzero(&newtio, sizeof(newtio));
    newtio.c_cflag = B115200 | CS8 | CLOCAL ;
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;
    newtio.c_lflag = 0;
    newtio.c_cc[VTIME] = 0;
    newtio.c_cc[VMIN] = 1;
    cfsetspeed(&newtio, B115200);
    tcflush(handle, TCIFLUSH);
    if (tcsetattr(handle,TCSANOW,&newtio))
    {
      perror("tcsetattr()");
      close(handle);
      return 1;
    }
#endif

  if (send_direction)  // downloading to NXT
  {
    searchhandle = (SEARCHHANDLETYPE)-1;
    do 
	{
      if (multiple)
	  {
		// first file?
		if (searchhandle == (SEARCHHANDLETYPE)-1) searchhandle = _findfirst32(local_filename, &finfo);
		else if (_findnext32(searchhandle, &finfo))
		{
		  // no more files found?
		  if (errno == ENOENT) 
		  {
		    _findclose(searchhandle);
		    break;
		  }
		}
		if (searchhandle == (SEARCHHANDLETYPE)-1)
		{
		  if (errno == ENOENT)
		  {
		    printf("No files found that match %s\n", local_filename);
			CloseHandle(handle);
		    return 1;
		  }
		  printf("Error searching directory %s, errno=%d\n", local_filename, errno);
		  CloseHandle(handle);
		  return 1;
		}

		// skip the directory entries
#ifdef WIN32
		if (finfo.attrib & _A_SUBDIR) continue;
		nm = finfo.name;
#else
		if (!finfo)
		{
		  _findclose(searchhandle);
		  break;
		}

		if (finfo->d_type != DT_REG) continue;  //assumes bsd-type structure
		nm = finfo->d_name;
#endif

		if (strlen(local_path) + strlen(nm) < 399)
		  sprintf(tmpname, "%s%s", local_path, nm);
		else
		{
		  printf("Too long filename/path %s%s\n", local_path, nm);
		  CloseHandle(handle);
		  return 1;
		}
        local_file = tmpname;
		nxt_file = nm;
	  }
	  else 
	  {
	    local_file = local_filename;
		nxt_file = nxt_filename;
	  }
	
	  // print info
	  printf("%s -> %s:%s ...", local_file, portname, nxt_file);
      fflush(stdout);

	  // open the local file
      FILE *f = fopen(local_file, "rb");
      if (!f)
      {
	    printf("Could not open the file to download %s, errno=%d\n", local_file, errno);
		CloseHandle(handle);
	    return 1;
      }
      len = _filelength(_fileno(f));
	  written = 0;
	
	  // delete the remote file, if it exists
	  d = (int)strlen(nxt_file);
	  p[0] = 3 + d;
	  p[1] = 0;
	  p[2] = 0x81;   // do not require reply from NXT
	  p[3] = 133;
	  strncpy((char *)p + 4, nxt_file, d + 1);
      send(p);

	  // determine the type of file (executables need to be written as linear)
	  linear = 0;
	  
	  if (d > 4) 
		if ((strcmp(nxt_file + d - 4, ".rxe") == 0) || 
            (strcmp(nxt_file + d - 4, ".RXE") == 0)) linear = 1;

	  // open NXT file
	  p[0] = 26;
	  p[1] = 0;
	  p[2] = 1;
	  if (linear) p[3] = 0x89;
	  else p[3] = 0x8b;
	  for (i = 4; i < 24; i++) p[i] = 0;
	  strncpy((char *)p + 4, nxt_file, d);
      p[24] = len & 0xFF;
	  p[25] = (len >> 8) & 0xFF;
	  p[26] = (len >> 16) & 0xFF;
	  p[27] = (len >> 24) & 0xFF;
	  send(p);

      //check if opening was successful
	  i = 0;
	  while (i < 6)
      {
	    int c = receive();
	    int tm = 0;
	    while ((c == -2) && (tm++ < 5))   // timeout, but accept only 15 times
	    {
              c = receive();
	      if (c==-2) Sleep(10);
	    }
        if (c < 0) 
		{
			printf("Error opening NXT file %s, open not confirmed\n", nxt_file);
			fclose(f);
			CloseHandle(handle);
	        return 1;
		}
	    nxt_msg[i++] = (unsigned char)c;
      }

	  // error opening file?
	  if (nxt_msg[4])  
	  {
		  printf("Error opening NXT file %s, err: %s\n", nxt_file, nxt_error(nxt_msg[4]));
		  fclose(f);
		  CloseHandle(handle);
	      return 1;
	  }

	  // NXT file open ok
	  nxthandle = nxt_msg[5];

	  // write the file in 48-byte chunks
	  while (len)
	  {
		// read and write next chunk 
		nread = (int)fread(buf, 1, 48, f);
		if (nread == 0)
		{
			if (ferror(f) || feof(f))
			{
			  printf("Error reading from file %s, errno=%d\n", local_file, errno);
			  fclose(f);
			  CloseHandle(handle);
			  return 1;
			}
		}
		len -= nread;
		written += nread;
		if (!len) skip_reply = 0; else skip_reply = 0x80;

		// write packet containing next chunk data
		p[0] = nread + 3;
		p[1] = 0;
		p[2] = 1 + skip_reply;
		p[3] = 0x83;  //write command
		p[4] = nxthandle;
		memcpy(p + 5, buf, nread);
		send(p);
	  }
	  fclose(f);

	  //check if write was successful
	  i = 0;
	  while (i < 8)
      {
	    int c = receive();
	    int tm = 0;
	    while ((c == -2) && (tm++ < 5))   // timeout, but accept only 5 times
          c = receive();
        if (c < 0) 
		{
			printf("Error writing NXT file %s, write not confirmed\n", nxt_file);
			CloseHandle(handle);
	        return 1;
		}
	    nxt_msg[i++] = (unsigned char)c;
      }

	  // error writing file?
	  if (nxt_msg[4])  
	  {
		printf("Error writing NXT file %s, err: %s\n", nxt_file, nxt_error(nxt_msg[4]));
		CloseHandle(handle);
        return 1;
	  }

	  // close the NXT file
	  p[0] = 3;
	  p[1] = 0;
	  p[2] = 0x01;   
	  p[3] = 0x84;
	  p[4] = nxthandle;
      send(p);

  	  //check if close was successful
	  i = 0;
	  while (i < 6)
      {
	    int c = receive();
	    int tm = 0;
	    while ((c == -2) && (tm++ < 5))   // timeout, but accept only 5 times
          c = receive();
        if (c < 0) 
		{
		  printf("Error closing NXT file %s, close not confirmed\n", nxt_file);
		  CloseHandle(handle);
	      return 1;
		}
	    nxt_msg[i++] = (unsigned char)c;
      }

	  // error closing file?
	  if (nxt_msg[4])  
	  {
		printf("Error closing NXT file %s, err: %s\n", nxt_file, nxt_error(nxt_msg[4]));
		CloseHandle(handle);
        return 1;
	  }
	  printf("%d bytes copied\n", written);

	} while (multiple);      // send next file, if multiple files requested
	
  }
  else // uploading from NXT
  {
    ffhandle = 0xFF; // first file
    do 
	{
      if (multiple)
	  {
	    if (ffhandle == 0xFF)  // first file
		{
	      // prepare NXT "find-first" message packet
          p[0] = 3 + (unsigned char)strlen(nxt_filename);
          p[1] = 0;
          p[2] = 1; // request reply
          p[3] = 134;    // message find first
		  strncpy((char *)p + 4, nxt_filename, 1 + strlen(nxt_filename));
          // send the prepared packet to the virtual serial port
          send(p);

          // retrieve packet with 30 bytes
          i = 0;
          while (i < 30) 
          {
	        int c = receive();
	        int tm = 0;
	        while ((c == -2) && (tm++ < 5)) // timeout, but accept only 5 times
            c = receive();
            if (c < 0) return -2;

	        nxt_msg[i++] = (unsigned char)c;
          }
          ffhandle = nxt_msg[5];

		  if (nxt_msg[4] == 135)
		  {
		    printf("No NXT files %s found\n", nxt_filename);
			CloseHandle(handle);
			return 1;
		  }

		  /* if the status is error, do not continue */
	      if (nxt_msg[4]) 
	      {
            printf("NXT responded to find first packet with error status, err:\n", nxt_error(nxt_msg[4]));
			CloseHandle(handle);
	        return 1;
	      }
		}
		else
		{
		  // prepare find next packet
	      p[0] = 3;
          p[1] = 0;
          p[2] = 1;      // request reply
          p[3] = 135;    // message find next
	      p[4] = ffhandle;
		  send(p);	

		  // retrieve packet with 30 bytes
          i = 0;
          while (i < 30) 
          {
	        int c = receive();
	        int tm = 0;
	        while ((c == -2) && (tm++ < 5)) // timeout, but accept only 5 times
              c = receive();
            if (c < 0) return -2;

	        nxt_msg[i++] = (unsigned char)c;
          }

		  if ((nxt_msg[4]) && (nxt_msg[4] != 135))
		  {
			printf("NXT responded to find next packet with err: %s\n", nxt_error(nxt_msg[4]));
			CloseHandle(handle);
	        return 1;
	      }

		  if (nxt_msg[4] == 135) break;
		}
        if (strlen(local_path) + strlen((char *)nxt_msg) < 399)
	      sprintf(tmpname, "%s%s", local_path, nxt_msg + 6);
	    else
	    {
	      printf("Too long filename/path %s%s\n", local_path, nxt_msg + 6);
		  CloseHandle(handle);
	      return 1;
	    }
        local_file = tmpname;
	    nxt_file = tmpname + strlen(local_path);
	  }
	  else 
	  {
	    nxt_file = nxt_filename;
		local_file = local_filename;
	  }
	
	  printf("%s:%s -> %s ...", portname, nxt_file, local_file);
	  fflush(stdout);

	  //nxt_file = "Forward.rso";

	  // open the local file
      FILE *f = fopen(local_file, "wb+");
      if (!f)
      {
	    printf("Could not open the file %s, errno=%d\n", local_file, errno);
		CloseHandle(handle);
	    return 1;
      }
	  
      // open NXT file for reading
	  p[0] = 22;
	  p[1] = 0;
	  p[2] = 1;
	  p[3] = 0x80;
	  for (i = 4; i < 24; i++) p[i] = 0;
	  d = (int)strlen(nxt_file);
	  strncpy((char *)p + 4, nxt_file, d + 1);
      send(p);

      //check if opening was successful
	  i = 0;
	  while (i < 10)
      {
	    int c = receive();
	    int tm = 0;
	    while ((c == -2) && (tm++ < 5))   // timeout, but accept only 5 times
	    {
              c = receive();
	      Sleep(10);
	    }
        if (c < 0) 
		{
			printf("Error opening NXT file %s, open not confirmed (%d,%d)\n", nxt_file, c, errno);
			fclose(f);
			CloseHandle(handle);
	        return 1;
		}
	    nxt_msg[i++] = (unsigned char)c;
      }

	  // error opening file?
	  if (nxt_msg[4])  
	  {
		  printf("Error opening NXT file %s, err: %s\n", nxt_file, nxt_error(nxt_msg[4]));
		  fclose(f);
		  CloseHandle(handle);
	      return 1;
	  }

	  // NXT file open ok
	  nxthandle = nxt_msg[5];
	  len = nxt_msg[6] + (nxt_msg[7] << 8) + (nxt_msg[8] << 16) + (nxt_msg[9] << 24);
	  written = 0;

	  // read the file in 48-byte chunks
	  while (len)
	  {
		nread = (len > 48) ? 48 : len;
		// request next packet containing next chunk data
		p[0] = 5;
		p[1] = 0;
		p[2] = 1;
		p[3] = 0x82;  //read command
		p[4] = nxthandle;  
		p[5] = nread & 0xFF;
        p[6] = nread >> 8;
		send(p);

		// retrive the data
	    i = 0;
	    while (i < 8 + nread)
        {
	      int c = receive();
	      int tm = 0;
	      while ((c == -2) && (tm++ < 5))   // timeout, but accept only 5 times
            c = receive();
          if (c < 0) 
		  {
			printf("Error writing NXT file %s, write not confirmed\n", nxt_file);
			CloseHandle(handle);
			fclose(f);
	        return 1;
		  }
	      nxt_msg[i++] = (unsigned char)c;
        }

		// error reading file?
	    if (nxt_msg[4])  
	    {
		  printf("Error reading NXT file %s, err: %s\n", nxt_file, nxt_error(nxt_msg[4]));
		  CloseHandle(handle);
		  fclose(f);
          return 1;
	    }

		// read all that we requested?
		if (nread > nxt_msg[6] + (nxt_msg[7] << 8))
		{
		  printf("Error reading NXT file %s, %d bytes received when %d requested\n", nxt_file, 
			     nxt_msg[6] + (nxt_msg[7] << 8), nread);
		  CloseHandle(handle);
		  fclose(f);
          return 1;
		}

		// write the data to file 
		nwrote = (int)fwrite(nxt_msg + 8, 1, nread, f);
		if (nwrote == 0)
		{
			if (ferror(f))
			{
			  printf("Error writing to file %s, errno=%d\n", local_file, errno);
			  CloseHandle(handle);
			  fclose(f);
			  return 1;
			}
		}
		len -= nread;
		written += nread;
	  }
	  fclose(f);

	  // error writing file?
	  if (nxt_msg[4])  
	  {
		printf("Error writing NXT file %s, err: %s\n", nxt_file, nxt_error(nxt_msg[4]));
		CloseHandle(handle);
        return 1;
	  }

	  // close the NXT file
	  p[0] = 3;
	  p[1] = 0;
	  p[2] = 0x01;   
	  p[3] = 0x84;
	  p[4] = nxthandle;
      send(p);

  	  //check if close was successful
	  i = 0;
	  while (i < 6)
      {
	    int c = receive();
	    int tm = 0;
	    while ((c == -2) && (tm++ < 5))   // timeout, but accept only 5 times
          c = receive();
        if (c < 0) 
		{
			printf("Error closing NXT file %s, close not confirmed\n", nxt_file);
			CloseHandle(handle);
	        return 1;
		}
	    nxt_msg[i++] = (unsigned char)c;
      }

	  // error closing file?
	  if (nxt_msg[4])  
	  {
		printf("Error closing NXT file %s, err: %s\n", nxt_file, nxt_error(nxt_msg[4]));
		CloseHandle(handle);
        return 1;
	  }
	  printf("%d bytes copied\n", written);
	} while (multiple); // send next file, if multiple files requested

  }

  CloseHandle(handle);
  return 0;

}
