//nxtrm.cpp - command line utility for removing files from NXT flash

#include <stdio.h>
#include <string.h>
#include <errno.h>

#ifdef WIN32
#include <windows.h>
#include <io.h>
#else
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <sys/types.h>
#include <fcntl.h>
#define HANDLE int
#define WriteFile(A,B,C,D,E) write(A,B,C)
#define ReadFile(A,B,C,D,E) read(A,B,C)
#define FlushFileBuffers(A) fsync(A)
#define CloseHandle close
#define DWORD unsigned int
#define INVALID_HANDLE_VALUE -1
#endif

static HANDLE handle;

// send NXT message string to port
void send(unsigned char *data)
{
  DWORD written;
  WriteFile(handle, data, (DWORD)(data[0] + 2), &written, 0);
  FlushFileBuffers(handle);
}

// receive one byte from port (returns after ca. 1.5 seconds if nothing received)
// returns -1 in case of error, -2 in case of timeout, otherwise the received char
int receive(void)
{
  unsigned char buf;
  DWORD nread;
  int rv = ReadFile(handle, &buf, 1, &nread, 0);
#ifdef WIN32
  if (rv == 0) return -1;
  if (nread == 0) return -2;
#else 
  if (rv == -1) return -1;
  if (rv == -2) return -2;
#endif
  return buf;
}

void print_usage()
{
	printf("\nnxtrm, remove files from NXT flash over BlueTooth virtual serial port\n");
	printf("\nusage: nxtrm nxt_port:filename\n\n");
}

char *nxt_error(int e)
{
	static char *nxterrs[] = { "No more handles", "No space", "No more files",
		                       "End of file expected", "End of file", "Not a linear file",
							   "File not found", "Handle all ready closed", "No linear space",
							   "Undefined error", "File is busy", "No write buffers",
							   "Append not possible", "File is full", "File exists",
							   "Module not found", "Out of boundary", "Illegal file name",
							   "Illegal handle" };
	if (!e) return "OK";
	if ((e > 0x80) && (e <= 0x93)) return nxterrs[e - 0x81];
	return "Unknown Error";
}

int main(int argc, char **argv)
{
  unsigned char p[100];
  unsigned char nxt_msg[100];
  char *filename;
  char portname[50];
  int d, i, ffhandle;

  //extract virtual port name from command line
  if (argc < 2)
  {
    print_usage();
	return 0;
  }

  if (strlen(argv[1]) < 5) 
  {
    print_usage();
#ifdef WIN32
	printf(" NXT port must be in the form COMx: or COMXX:\n");
#else
	printf(" NXT port must be in the form /dev/rfcommX:\n");
#endif
  }

  if (argv[1][4] == ':') 
  {
    strncpy(portname, argv[1], 4);
	portname[4] = '\0';
	filename = argv[1] + 5;
  }
  else
#ifdef WIN32
  {
	strncpy(portname, "\\\\.\\", 4);
	strncpy(portname + 4, argv[1], 5);
	portname[9] = '\0';
	filename = argv[1] + 6;
  }
#else
  {
        if (strlen(argv[1]) < 13) { printf("Do not understand %s\n", argv[1]); return 1; }
        int pos; 
	if (argv[1][12] == ':') pos = 12;
	else if (argv[1][13] == ':') pos = 13;
        else { printf("Do not understand %s\n", argv[1]); return 1; }
        strncpy(portname, argv[1], pos);
	portname[pos] = '\0';
	filename = argv[1] + pos + 1;
  }
#endif

  // open the NXT serial port
#ifdef WIN32
  handle = CreateFileA(portname, FILE_READ_DATA | FILE_APPEND_DATA | FILE_WRITE_DATA, 0, NULL, OPEN_EXISTING, 0, NULL);
  if (handle != INVALID_HANDLE_VALUE) 
  {
    SetupComm(handle, 90, 90);

    COMMTIMEOUTS timeouts;
    timeouts.ReadIntervalTimeout = 0;
    timeouts.ReadTotalTimeoutConstant = 0;
    timeouts.ReadTotalTimeoutMultiplier = 1500;  //1.5s timeout on read() from serial port
    timeouts.WriteTotalTimeoutMultiplier = 0;
    timeouts.WriteTotalTimeoutConstant = 1500;  //1.5s timeout on write() to serial port
    SetCommTimeouts(handle, &timeouts);

    PurgeComm(handle, PURGE_RXCLEAR | PURGE_TXCLEAR);
  }
  else 
  {
    printf("Could not open the port %s, err=%d\n\n", portname, GetLastError());
    return 1;
  }
#else
    struct termios newtio, oldtio;
    handle = open(portname, O_RDWR | O_NOCTTY);
    if (handle < 0) { perror(portname); return INVALID_HANDLE_VALUE; }

    tcgetattr(handle,&oldtio); /* save current port settings */

    bzero(&newtio, sizeof(newtio));
    newtio.c_cflag = B115200 | CS8 | CLOCAL ;
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;
    newtio.c_lflag = 0;
    newtio.c_cc[VTIME] = 0;
    newtio.c_cc[VMIN] = 1;
    cfsetspeed(&newtio, B115200);
    tcflush(handle, TCIFLUSH);
    if (tcsetattr(handle,TCSANOW,&newtio))
    {
      perror("tcsetattr()");
      close(handle);
      return 1;
    }
#endif

  // check if the file name contains wildcards
  if (strchr(filename, '*'))
  {
    // need to use findfirst/findnext to retrieve names of files to delete
  
    // prepare NXT "find-first" message packet
    p[0] = (unsigned char)(3 + strlen(filename));
    p[1] = 0;
    p[2] = 1; // request reply
    p[3] = 134;    // message find first
	strncpy((char *)p + 4, filename, strlen(filename) + 1);
    // send the prepared packet to the virtual serial port
    send(p);

    // retrieve packet with 30 bytes
    i = 0;
    while (i < 30)
    {
	  int c = receive();
	  int tm = 0;
	  while ((c == -2) && (tm++ < 5)) // timeout, but accept only 5 times
        c = receive();
      if (c < 0) return -2;

	  nxt_msg[i++] = (unsigned char)c;
    }

    ffhandle = nxt_msg[5];
	
	if (nxt_msg[4] == 135)
	{
	  printf("No files to delete found\n");
	  CloseHandle(handle);
	  return 1;
	}

	/* if the status is error, do not continue */
	if (nxt_msg[4])
	{
	  printf("NXT responded to find first packet with error status, err=%d\n", nxt_msg[4]);
	  CloseHandle(handle);
	  return 1;
	}

	do {    
	  // delete the file
      d = (int)strlen((char *)nxt_msg + 6);
      p[0] = 3 + d;
      p[1] = 0;
      p[2] = 1;   
      p[3] = 133;
      strncpy((char *)p + 4, (char *)nxt_msg + 6, d + 1);
      send(p);

      //check if delete was successful
      i = 0;
      while (i < 25)
      {
        int c = receive();
        int tm = 0;
        while ((c == -2) && (tm++ < 5))   // timeout, but accept only 5 times
          c = receive();
        if (c < 0) 
	    {
          printf("Error deleting NXT file %s, delete not confirmed\n", p + 4);
	      CloseHandle(handle);
	      return 1;
	    }
        nxt_msg[i++] = (unsigned char)c;
      }

      // error deleting file?
      if (nxt_msg[4])  
      {
        printf("Error deleting NXT file %s, err: %s\n", p + 4, nxt_error(nxt_msg[4]));
        CloseHandle(handle);
        return 1;
      }

      printf("%s deleted\n", p + 4);

 	  // prepare find next packet
	  p[0] = 3;
      p[1] = 0;
      p[2] = 1;      // request reply
      p[3] = 135;    // message find next
	  p[4] = ffhandle;
	  // and send it
	  send(p);

      // retrieve packet with 30 bytes
      i = 0;
      while (i < 30)
      {
	    int c = receive();
	    int tm = 0;
	    while ((c == -2) && (tm++ < 5)) // timeout, but accept only 5 times
          c = receive();
        if (c < 0) return -2;

	    nxt_msg[i++] = (unsigned char)c;
      }

	} while (!nxt_msg[4]);

	/* if the status is error, do not continue */
	if (nxt_msg[4] != 135)
	{
	  printf("NXT responded to find next packet with error status\n");
	  return 1;
	}
  }
  else  // only single file deletion
  {
    // delete the remote file, if it exists
    d = (int)strlen(filename);
    p[0] = 3 + d;
    p[1] = 0;
    p[2] = 1;   
    p[3] = 133;
    strncpy((char *)p + 4, filename, d + 1);
    send(p);

    //check if delete was successful
    i = 0;
    while (i < 25)
    {
      int c = receive();
      int tm = 0;
      while ((c == -2) && (tm++ < 5))   // timeout, but accept only 5 times
        c = receive();
      if (c < 0) 
	  {
        printf("Error deleting NXT file %s, delete not confirmed\n", filename);
	    CloseHandle(handle);
	    return 1;
	  }
      nxt_msg[i++] = (unsigned char)c;
    }

    // error deleting file?
    if (nxt_msg[4])  
    {
      printf("Error deleting NXT file %s, err: %s\n", filename, nxt_error(nxt_msg[4]));
      CloseHandle(handle);
      return 1;
    }

    printf("%s deleted\n", filename);
  }
  CloseHandle(handle);
  return 0;
}
