// remote1.cpp: sends one-line expression to NXT Logo and retrieves the answer
//
// Description:
// The program reads NXT Logo expression from 
// standard input, transmits it over BlueTooth virtual serial 
// port to NXT, waits for a resulting expression arriving
// from NXT, and prints the resulting expression to the standard output.
// The complete expression must be contained at 
// a single line. Multi-line expressions (such as Logo programs) 
// stored in a separate file can be sent using command 
// 'include file_name'. All resulting expressions are logged
// to file remote.log. 
//
// usage:  remote1 virtual_port_name

#ifdef WIN32
#include <windows.h>
#else
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#define HANDLE int
#define INVALID_HANDLE_VALUE -1
#define WriteFile(A,B,C,D,E) write(A,B,C)
#define ReadFile(A,B,C,D,E) read(A,B,C)
#define FlushFileBuffers(A) fsync(A)
#define DWORD unsigned int
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>

#define DEFAULT_LOG_FILE "remote.log"
#define NXTLOGO_MAX_EXPRESSION_LEN 10000


static HANDLE handle;  // virtual serial port handle

// send NXT message string to port
void send(unsigned char *data)
{
  DWORD written;
  WriteFile(handle, data, (DWORD)(data[0] + 2), &written, 0);
  FlushFileBuffers(handle);
}

// send a message packet of a specified length to the specified mailbox of NXT
void send_message(unsigned char *data, int len, int inbox)
{
  // prepare NXT message packet
  unsigned char *p = (unsigned char *)malloc(len + 7);
  p[0] = len + 5;
  p[1] = 0;
  p[2] = 0x80; // request no reply
  p[3] = 9;    // send message direct command
  p[4] = inbox; // target mailbox
  p[5] = len + 1; // length of the message
  memcpy(p + 6, data, len);
  p[len + 6] = 0; // terminate message with 0
  
  // send the prepared packet to the virtual serial port
  send(p);
}

// receive one byte from port (returns after ca. 1.5 seconds if nothing received)
// returns -1 in case of error, -2 in case of timeout, otherwise the received char
int receive(void)
{
  unsigned char buf;
  DWORD nread;
  int rv = ReadFile(handle, &buf, 1, &nread, 0);
  if (rv == 0) return -1;
  if (nread == 0) return -2;
  return buf;
}

// receive a NXT message from virtual serial port
// returns -2 in case of error, -1 if there is no message available, or NXT reports another error,
// otherwise the length of the message is returned
// if message is received, s will point to the message data (a static string that should be copied 
//  away before calling receive_message() again, and should not be deallocated 
//  - i.e. this call is not multiple-threads safe!)
int receive_message(unsigned char **s, int remote_inbox, int local_inbox, int remove_msg)
{
  // message request packet
  static unsigned char msg_request[] = { (unsigned char)5, (unsigned char)0, (unsigned char)0, (unsigned char)19, 
	                                    (unsigned char)0xFF, (unsigned char)0xFF, (unsigned char)0xFF };
  static unsigned char nxt_msg[66];
  int i;

  // since PC is master, we must request a reply message containing 
  // a message that may be pending at NXT slave

  // prepare NXT request message packet
  msg_request[4] = remote_inbox;
  msg_request[5] = local_inbox;
  msg_request[6] = remove_msg;

  // send the prepared packet to the virtual serial port
  send(msg_request);

  // retrieve packet with 66 bytes
  i = 0;
  while (i < 66) 
  {
	  int c = receive();
	  int tm = 0;
	  while ((c == -2) && (tm++ < 5)) // timeout, but accept only 5 times
        c = receive();
      if (c < 0) return -2;

	  nxt_msg[i++] = (unsigned char)c;
  }

  // reply packet reports error - probably no message available
  if (nxt_msg[4] != 0) return -1;

  // real message data start at index 7
  *s = nxt_msg + 7;

  return nxt_msg[6];  // message length is at index 6
}

// send a string expression packed into one or more packets to NXT Logo
// returns -1 if the expression is too long, otherwise returns 0
int send_expression(unsigned char *e, int len)
{
  unsigned char *p, *r, *confirm;
  int l, waitconf;

  if (len + 1 > NXTLOGO_MAX_EXPRESSION_LEN) return -1;
  p = (unsigned char *)malloc(58);

  // prepare NXT Logo packet: 
  // the first 2 bytes of the first packet contain the expression length
  p[0] = len >> 8;
  p[1] = len & 0xFF;
  // first packet has at most 56 bytes of the expression
  l = (len < 56) ? len : 56;
  memcpy(p + 2, e, l);
  len -= l;
  l += 2;      // the first packet contains extra two bytes with the length of the expression

  r = e + 56;   // pointer to remaining data
  waitconf = 0; // do not wait for confirmation before first packet

  while (l)  // while there are data to be sent
  {
	if (waitconf) // wait for confirmation message from NXT Logo, unless it is the first packet
	  while (-1 == receive_message(&confirm, 3, 3, 1));

	send_message(p, l, 1);

	if (!waitconf)
	{
	  waitconf = 1; // wait for confirmation next time before the next packet is sent 
      free(p);
	}

	// extract next chunk if data remaining
    l = (len < 56) ? len : 56;
    len -= l;
    p = r;
    r = p + 56;   // pointer to remaining data
  }
  return 0;
}

// receive reply expression from NXT Logo
// returns -1 in case of error, otherwise returns length of the retrieved expression
// e will point to newly allocated string, the caller is responsible to deallocate it
int receive_expression(unsigned char **e)
{
  unsigned char *m, *wp, *limit;
  int l, len;
  
  // wait for NXT to evaluates the expression and sends us the reply expression 
  while (-1 == (l = receive_message(&m, 2, 2, 1)));

  // if error received, return directly
  if (l == -2) return -1;

  // should have received at least 2 bytes
  if (l < 2) return -1;

  // retrieve the length of the result expression
  len = (m[0] << 8) + m[1];

  // allocate the result expression
  *e = (unsigned char *) malloc(len + 1);
  limit = (*e) + len;
  (*e)[len] = 0;  // add terminating 0

  // the effective length in the first packet
  l = (len < 56) ? len : 56;
  len -= l;

  // retrieve the first part of the expression
  memcpy(*e, m + 2, l);
  wp = (*e) + l;   // next chunk write pointer

  // retrieve remaining part of the expression if any
  while (len > 0)
  {
    // request more content
    send_message((unsigned char *)"+", 1, 4);
    
	// and retrieve it
    while (-1 == (l = receive_message(&m, 2, 2, 1)));
 
	// if error, stop receiving process, alse make sure we do not go out of bounds
	if ((l == -2) || (wp + l > limit))
	{
		free(*e);
		return -1;
	}

	// copy new chunk to the target and advance write pointer
	memcpy(wp, m, l);
	wp += l;

	// determine how much remains
    l = (len < 56) ? len : 56;
    len -= l;
  }

  return (int)(limit - (*e));
}

//program entry point
int main(int argc, char **argv)
{
	char portname[15];
	unsigned char expression[NXTLOGO_MAX_EXPRESSION_LEN + 1];
	unsigned char *s, *reply;
	int i;
	char c;

	//extract virtual port name from command line
	if (argc != 2)
	{
		printf("usage: remote virtual_port_name\n");
		return 0;
	}

	int arglen = (int)strlen(argv[1]);
#ifdef WIN32
	if (arglen > 10) 
	{
		printf("\n# Port name %s cannot be accepted.\n\n", argv[1]);
		return 1;
	}
#endif


#ifdef WIN32
	if ((strlen(argv[1]) != 4) && (argv[1][0] != '\\'))
	{
		strncpy(portname, "\\\\.\\", 4);
		strncpy(portname + 4, argv[1], arglen + 1);
	}
	else 
#endif
	  strncpy(portname, argv[1], arglen + 1);

	// open the port
#ifdef WIN32
    handle = CreateFileA(portname, FILE_READ_DATA | FILE_APPEND_DATA | FILE_WRITE_DATA, 0, NULL, OPEN_EXISTING, 0, NULL);
    if (handle != INVALID_HANDLE_VALUE) 
    {
	  SetupComm(handle, 90, 90);

	  COMMTIMEOUTS timeouts;
	  timeouts.ReadIntervalTimeout = 0;
	  timeouts.ReadTotalTimeoutConstant = 0;
	  timeouts.ReadTotalTimeoutMultiplier = 1500;  //1.5s timeout on read() from serial port
	  timeouts.WriteTotalTimeoutMultiplier = 0;
	  timeouts.WriteTotalTimeoutConstant = 1500;  //1.5s timeout on write() to serial port
	  SetCommTimeouts(handle, &timeouts);

      PurgeComm(handle, PURGE_RXCLEAR | PURGE_TXCLEAR);
    }
    else 
    {
	  printf("# Could not open the port, err=%d\n\n", GetLastError());
	  return 1;
    }
#else
    struct termios newtio, oldtio;
    handle = open(portname, O_RDWR | O_NOCTTY);
    if (handle < 0) { perror(portname); return INVALID_HANDLE_VALUE; }

    tcgetattr(handle,&oldtio); /* save current port settings */

    bzero(&newtio, sizeof(newtio));
    newtio.c_cflag = B115200 | CS8 | CLOCAL ;
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;
    newtio.c_lflag = 0;
    newtio.c_cc[VTIME] = 0;
    newtio.c_cc[VMIN] = 1;
    cfsetspeed(&newtio, B115200);
    tcflush(handle, TCIFLUSH);
    if (tcsetattr(handle,TCSANOW,&newtio))
    {
      perror("tcsetattr()");
      close(handle);
      return 1;
    }
#endif

    do {

	// open output log file
	FILE *f = fopen(DEFAULT_LOG_FILE, "a+");
    if (f == 0)
		printf("# Could not open log file err=%d, logging will be disabled.\n\n", errno);

	  // read expression from standard input
	  i = 0;
	  do {
	    scanf("%c", &c);
	    if (c != '\n') expression[i++] = (unsigned char)c;
	  } while ((c != '\n') && (i < NXTLOGO_MAX_EXPRESSION_LEN));
	  expression[i] = 0;
	  
	  // check for include command
	  if (strncmp("include", (char *)expression, 7) == 0)
	  {
		  // read the whole expression from file
		  s = expression + 7;
		  while ((*s == ' ') || (*s == '\t')) s++;
		  FILE *g = fopen((char *)s, "r");
		  if (g == 0) 
		  {
			  printf("\n# Could not open include file, err=%d\n\n", errno);
			  break;
		  }

		  i = 0;
		  while (EOF != fscanf(g, "%c", &c)) expression[i++] = c;
		  expression[i] = 0;
	  }
	  else i = (int)strlen((char *)expression);

	  // write it to the log file
	  if (f) fprintf(f, "-E-\n%s\n-R-\n", expression);
	  
	  // send the expression to NXT
	  if (-1 == send_expression(expression, i))
		  printf("# Expression is too long.\n");

	  // receive reply expression
	  receive_expression(&reply);

	  // print the expression 
	  printf("%s\n", reply);

	  // write it to the log file
	  if (f) fprintf(f, "%s\n", reply);
	
	// close output log file
	if (f) fclose(f);

    } while (0);

	// close the port
#ifdef WIN32
    PurgeComm(handle, PURGE_RXCLEAR | PURGE_TXCLEAR);
    CloseHandle(handle);
#else
    close(handle);
#endif
	return 0;
}
